module JsonRpc exposing
    ( WebData
    , Response(..), RpcError, RpcData
    , TaskData, TaskRpcData
    , Data(..)
    , Param
    , call, callWithConfig
    , callTask, callWithConfigTask
    , flatResponse, flat, toResult, httpErrToString
    )

{-| 一个通用的JsonRpc-V2调用接口，帮助实现JsonRpc客户端应用。


# 数据结构


## Http 数据结构

@docs WebData


## JsonRpc 数据结构

@docs Response, RpcError, RpcData


## 用于 Task 调用的 JsonRpc 数据结构

@docs TaskData, TaskRpcData


## 便于用户使用的 JsonRpc 数据结构

@docs Data


## JsonRpc 调用参数结构

@docs Param


# Cmd 方式调用

@docs call, callWithConfig


# Task 方式调用

@docs callTask, callWithConfigTask


# 辅助函数

@docs flatResponse, flat, toResult, httpErrToString

-}

import Http exposing (Header)
import Json.Decode as D
import Json.Decode.Pipeline exposing (optional, required)
import Json.Encode as E
import Task exposing (Task)


{-| JsonRpc 返回错误格式定义
-}
type alias RpcError =
    { code : Int
    , message : String
    , data : Maybe String
    }


{-| JsonRpc 响应结果
-}
type Response a
    = InnerResult a
    | InnerError RpcError


{-| Http 返回数据定义
-}
type alias WebData a =
    Result Http.Error a


{-| 基于 http 的 JsonRpc 响应结果
-}
type alias RpcData a =
    WebData (Response a)


{-| Http Task 调用返回数据定义
-}
type alias TaskData a =
    Task Http.Error a


{-| 基于 http Task 调用 JsonRpc 返回结果
-}
type alias TaskRpcData a =
    TaskData (Response a)


{-| 便于处理的返回结果
-}
type Data a
    = RpcResult a
    | RpcErr RpcError
    | HttpErr Http.Error


{-| 调用参数
-}
type alias Param =
    { url : String
    , token : Maybe String
    , method : String
    , params : List ( String, E.Value )
    }


{-| 将rpc调用的返回结果平面化处理
-}
flat : RpcData a -> Data a
flat rpc =
    case rpc of
        Ok d ->
            flatResponse d

        Err e ->
            HttpErr e


{-| 返回结果平面化处理
-}
flatResponse : Response a -> Data a
flatResponse rpc =
    case rpc of
        InnerResult v ->
            RpcResult v

        InnerError e ->
            RpcErr e


{-| 将平面化处理结果转换为Result
-}
toResult : Data a -> Result String a
toResult data =
    case data of
        RpcResult d ->
            Ok d

        RpcErr e ->
            Err <| errorToString e

        HttpErr he ->
            Err <| httpErrToString he


{-| Http.Error转换为String
-}
httpErrToString : Http.Error -> String
httpErrToString he =
    case he of
        Http.BadUrl u ->
            "无效URL => " ++ u

        Http.Timeout ->
            "超时"

        Http.NetworkError ->
            "网络出错"

        Http.BadStatus status ->
            "返回状态码" ++ String.fromInt status

        Http.BadBody reason ->
            reason


{-| 定义Json数据头
-}
acceptJson : Header
acceptJson =
    Http.header "Accept" "application/json"


{-| 定义认证头
-}
authHeader : String -> Header
authHeader token =
    Http.header "Authorization" ("bearer " ++ token)


{-| -}
type alias Config =
    { headers : List Header
    , timeout : Maybe Float
    , tracker : Maybe String
    }


{-| 缺省配置

  - accept application/json
  - no timeout
  - no tracker

-}
defaultConfig : Config
defaultConfig =
    { headers = [ acceptJson ]
    , timeout = Nothing
    , tracker = Nothing
    }


authConfig : String -> Config
authConfig token =
    { defaultConfig | headers = authHeader token :: defaultConfig.headers }



-- Request


{-| 内部调用的请求命令
-}
request : String -> Config -> String -> (WebData a -> msg) -> D.Decoder a -> E.Value -> Cmd msg
request method config url toMsg decoder body =
    Http.request
        { method = method
        , headers = config.headers
        , url = url
        , body = Http.jsonBody body
        , expect = Http.expectJson toMsg decoder
        , timeout = config.timeout
        , tracker = config.tracker
        }


{-| `POST`
-}
postWithConfig :
    Config
    -> String
    -> (WebData a -> msg)
    -> D.Decoder a
    -> E.Value
    -> Cmd msg
postWithConfig =
    request "POST"


{-| `POST` 缺省配置
-}
post :
    String
    -> (WebData a -> msg)
    -> D.Decoder a
    -> E.Value
    -> Cmd msg
post =
    postWithConfig defaultConfig


{-| `POST` 带Token
-}
postWithToken :
    Maybe String
    -> String
    -> (WebData a -> msg)
    -> D.Decoder a
    -> E.Value
    -> Cmd msg
postWithToken token =
    case token of
        Just t ->
            postWithConfig <| authConfig t

        Nothing ->
            postWithConfig <| defaultConfig


{-| HTTP调用方法，供定制扩展使用，如Zabbix调用。
-}
callWithConfig :
    (Param -> E.Value)
    -> Param
    -> D.Decoder a
    -> (RpcData a -> msg)
    -> Cmd msg
callWithConfig builder opts decoder toMsg =
    let
        decoderX =
            D.oneOf
                [ D.map InnerResult (D.at [ "result" ] decoder)
                , D.map InnerError (D.at [ "error" ] errorDecoder)
                ]
    in
    post opts.url toMsg decoderX (builder opts)


{-| HTTP调用方法
-}
callWithToken :
    (Param -> E.Value)
    -> Param
    -> D.Decoder a
    -> (RpcData a -> msg)
    -> Cmd msg
callWithToken builder opts decoder toMsg =
    let
        decoderX =
            D.oneOf
                [ D.map InnerResult (D.at [ "result" ] decoder)
                , D.map InnerError (D.at [ "error" ] errorDecoder)
                ]
    in
    postWithToken opts.token opts.url toMsg decoderX (builder opts)


{-| HTTP 调用方法
-}
call :
    Param
    -> D.Decoder a
    -> (RpcData a -> msg)
    -> Cmd msg
call =
    callWithToken jsonrpc_request



-- Task


{-| 内部调用的任务命令
-}
task : String -> Config -> String -> D.Decoder a -> E.Value -> TaskData a
task method config url decoder body =
    Http.task
        { method = method
        , headers = config.headers
        , url = url
        , body = Http.jsonBody body
        , resolver = Http.stringResolver <| handleJsonResponse <| decoder
        , timeout = config.timeout
        }


{-| `POST` request as a command, with additional `Config`.
postWithConfig defaultConfig == post
-}
postWithConfigTask :
    Config
    -> String
    -> D.Decoder a
    -> E.Value
    -> TaskData a
postWithConfigTask =
    task "POST"


{-| `POST` request as a command.
-}
postTask :
    String
    -> D.Decoder a
    -> E.Value
    -> TaskData a
postTask =
    postWithConfigTask defaultConfig


{-| `POST` 带Token
-}
postWithConfigAndTokenTask :
    Maybe String
    -> String
    -> D.Decoder a
    -> E.Value
    -> TaskData a
postWithConfigAndTokenTask token =
    case token of
        Just t ->
            postWithConfigTask <| authConfig t

        Nothing ->
            postWithConfigTask defaultConfig


{-| HTTP Task 调用方法。扩展zabbix接口时，可以用这个函数。
-}
callWithConfigTask :
    (Param -> E.Value)
    -> Param
    -> D.Decoder a
    -> TaskRpcData a
callWithConfigTask builder opts decoder =
    let
        decoderX =
            D.oneOf
                [ D.map InnerResult (D.at [ "result" ] decoder)
                , D.map InnerError (D.at [ "error" ] errorDecoder)
                ]
    in
    postTask opts.url decoderX (builder opts)


{-| HTTP Task 调用方法
-}
callWithTokenTask :
    (Param -> E.Value)
    -> Param
    -> D.Decoder a
    -> TaskRpcData a
callWithTokenTask builder opts decoder =
    let
        decoderX =
            D.oneOf
                [ D.map InnerResult (D.at [ "result" ] decoder)
                , D.map InnerError (D.at [ "error" ] errorDecoder)
                ]
    in
    postWithConfigAndTokenTask opts.token opts.url decoderX (builder opts)


{-| HTTP Task调用方法
-}
callTask :
    Param
    -> D.Decoder a
    -> TaskRpcData a
callTask =
    callWithTokenTask jsonrpc_request



-- Helper


{-| 构建JsonRpc请求体
-}
jsonrpc_request : Param -> E.Value
jsonrpc_request opts =
    E.object
        [ ( "id", E.int 0 )
        , ( "jsonrpc", E.string "2.0" )
        , ( "method", E.string opts.method )
        , ( "params", E.object opts.params )
        ]


{-| -}
handleJsonResponse : D.Decoder a -> Http.Response String -> WebData a
handleJsonResponse decoder response =
    case response of
        Http.BadUrl_ url ->
            Err (Http.BadUrl url)

        Http.Timeout_ ->
            Err Http.Timeout

        Http.BadStatus_ { statusCode } _ ->
            Err (Http.BadStatus statusCode)

        Http.NetworkError_ ->
            Err Http.NetworkError

        Http.GoodStatus_ _ body ->
            case D.decodeString decoder body of
                Err _ ->
                    Err (Http.BadBody body)

                Ok result ->
                    Ok result


{-| 错误格式解码器
-}
errorDecoder : D.Decoder RpcError
errorDecoder =
    D.succeed RpcError
        |> required "code" D.int
        |> required "message" D.string
        |> optional "data" (D.nullable D.string) Nothing


{-| -}
errorToString : RpcError -> String
errorToString e =
    "Code: "
        ++ String.fromInt e.code
        ++ " Message: "
        ++ e.message
        ++ (case e.data of
                Just d ->
                    " Data: " ++ d

                _ ->
                    ""
           )
